####################################################
##                                                ##
##    Model selection in the linear regression    ##
##                                                ##
####################################################

br = read.table("br.csv", sep = ",", header=T) # reads the data from a csv file   
br$Occupancy = as.factor(br$Occupancy)
br$Style = as.factor(br$Style)
attach(br)

## 0. ################################################
## Sample split

# We now partition our dataset in two samples.
# We randomly select a training sample of N = 600 observations.
Ntot = nrow(br)  # total n. of obs.
set.seed(123)    # random seed: it allows to replicate the analysis and obtain the same results
N = 0.6 * Ntot          # size of training sample
s = sample(1:Ntot, N)  # index of selected units
train = br[s,]   # training sample
test = br[-s,]   # validation sample
summary(train)   # data set summary 


# regression model formula: full model
model_full = price ~ sqft + I(sqft ^ 2) + I(sqft ^ 3) + Age + I(Age ^ 2 ) + 
  I(Age ^ 3) + I(sqft * Age) + I(sqft ^ 2 * Age) + I(sqft * Age ^ 2) +
  Pool + Baths + I(Baths ^ 2) + I(Baths ^ 3) + Bedrooms + I(Bedrooms ^ 2) + 
  I(Bedrooms ^ 3) + Fireplace + Waterfront + DOM + factor(Occupancy) + factor(Style) 
# Notice that we train the model only on the training sample
regr_full = lm(model_full, data = train)  
summary(regr_full)

## 1. ################################################
## Subset selection
p.max = 600 - df.residual(regr_full) # number of variables in the full model (remember factors' treatment)

# installing and/or importing the package 'leaps' that
# contains the function regsubsets()
if(!require(leaps)){
  install.packages("leaps")
  library(leaps) 
}

## Best subset selection
subr = regsubsets(model_full, 
                  data = train,
                  method = "forward",
                  nbest = 1,
                  nvmax = 25,
                  force.in = 1) # all the models will contain sqft (1st variable)
                                # if you set it to zero, sqft will be subject to selection 
subr.out = summary(subr) # this object will contain the subr output
names(subr.out) # we can see which objects are accessible in the output using $

subr.out$c # for example, Mallow's C_ps
min(subr.out$cp)           # minimum value 

# Comparing the different criteria by visualizing them in tabular form
cbind(subr.out$cp, subr.out$adjr2, 
      subr.out$bic, subr.out$rsq) 

# Extracting the indices corresponding to the minimum values of the information
# criteria criteria and store them for future use (one by one).
imin.cp = which(subr.out$cp == min(subr.out$cp))
imin.bic = which(subr.out$bic == min(subr.out$bic)) 
imax.adjr2 = which(subr.out$adjr2 == max(subr.out$adjr2)) 

# Comparing the different criteria by visualizing them in plots
# where it is also reported the previously stored index corresponding
# to minimum values
par(mfrow = c(3, 1)) # it allows to add several plots into a grid of dimension 3 row x 1 col
                     # it has to be turned off (see later)
plot(subr.out$cp, xlab = "Number of Variables", ylab = "Cp")
abline(v = imin.cp, pch = 20, col = "red")
plot(subr.out$bic, xlab = "Number of Variables", ylab = "BIC")
abline(v = imin.bic, pch = 20, col = "red")
plot(subr.out$adjr2, xlab = "Number of Variables", ylab = "Adj. R sq.")
abline(v = imax.adjr2, pch = 20, col = "red")
par(mfrow = c(1, 1))  # it sets back the grid for plots to 1 x 1

plot(subr, scale = "bic")
# each row is a model. Each shaded rectangle indicates whether
# the variable (column) is included in model (row). The value of Cp is 
# reported on the y axis, and is also indicated by the color intensity. 

subr.out$which # matrix indicating which variables are included in each model
subr.out$which[imin.bic, ] # row corresponding to the selected model (according to Cp)
subr$xnames[subr.out$which[imin.bic,]]# printing the variables of the model selected by the Cp criterion
coef(subr, imin.cp) # OLS coefficients for the selected model

# Estimate the linear model with the selected variables
model_sel = price ~ sqft + I(sqft ^ 2) + I(sqft ^ 3) + I(sqft  * Age) + I(sqft ^ 2 * Age) +
  I(sqft * Age ^ 2) + Baths + I(Baths ^ 2) + Bedrooms + I(Bedrooms^2) + I(Bedrooms ^ 3) +
  Fireplace + Waterfront + I(factor(Occupancy) == 2) + I(factor(Style) == 2) +
  I(factor(Style) == 4) + I(factor(Style) == 7) + I(factor(Style) == 10)

regr_sel = lm(model_sel, data = train)
summary(regr_sel)

## Backward and forward selection
# Try to apply forward subset selection and backward subset selection
# by changing the argument method = "exhaustive" in the function
# regsubsets() to "forward" or "backward", respectively. 
# Is the resulting outcome any different?

## 2. ################################################
## Validation: test sample prediction accuracy

## We can use the R function predict 
ypred_full = predict(regr_full, newdata = test) 
pred_errors_full = test$price - ypred_full; # full model predictions for the test sample
mse_full = mean(pred_errors_full  ^ 2)   # mean square prediction error of full model

ypred_sel = predict(regr_sel, newdata = test)
pred_errors_sel = test$price - ypred_sel;   # selected model predictions
mse_sel = mean(pred_errors_sel ^ 2)

100 * (1 - mse_sel/mse_full)  # percent gain in accuracy due to model selection
